/* /usr/X11R6/include/X11/Xlib.h */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include "kwav.h"
#include "kfft.h"

#define L 1000 /* fenetre mere */
#define H 700  /* fenetre mere */
#define M 10   /* marges */

#define L_BTN 43 /* fenetres boutons  */
#define H_BTN 30 /* fenetres boutons  */

#define H_PAN (1*H/12 - M)   /* panorama */
#define H_SEL (4*H/12)       /* selection */
#define H_SNG (6*H/12 - M)   /* sonogramme */
#define H_FFT (6*H/12 - M)   /* fft */
#define H_INF (H_PAN+H_SEL+M)/* info */

#define L_PAN (2*L/3 - M*2)  /* panorama */
#define L_SEL (2*L/3 - M*2)  /* selection */
#define L_SNG (2*L/3 - M*2)  /* sonogramme */
#define L_FFT (1*L/3 - M)    /* fft */
#define L_INF (1*L/3 - M)    /* info */

#define SIZE_NOM 256

Display *dpy;
GC gc;
GC gcFont;
XFontStruct *fd; /* la police */
XGCValues gcv;
Window racine, winMere, winPanorama, winSelection, winSono, winFft, winInfo, winSaisie;
Window winBLoad, winBPlay, winBRec, winBQuit, winBFft, winBSave, winBHead;

XEvent event;
XGCValues xgcv;
XGCValues gcv;
XSetWindowAttributes xswa;
XColor xcolor;
XpmAttributes attributes;
Pixmap pixmap;
int ecran;
long blanc;
long noir;
long bleu;
long rouge;
long vertFonce;
long gris;
long jaune;
long violet;
long vert;
long couleur;

kwav_struct sonwav;
kwav_struct selwav;
char nom[SIZE_NOM];

long conversion(int r,int v,int b);
void separation(long couleur, int *r, int *v, int *b);
unsigned int getHeight(Window win);
unsigned int getWidth(Window win);
void Xor(Window winxor);
void InitColor();
void InitWindow (Window *win, Window *mere, int posx, int posy, int largeur, int hauteur, int frontC, int backC, long event_mask);
void InitButton (char* filename, char* nomBouton, Window *win, Window *mere, int posx, int posy, int largeur, int hauteur, int frontC, int backC, long event_mask);
void InitWindows(int posx, int posy, int largeur, int hauteur);
void InitX11(int posx, int posy, int largeur, int hauteur);
void Dessine_wav(Window win, kwav_struct sonwav);
void Dessine_fft(Window win, kwav_struct sonwav);
void Dessine_sono(Window win, kwav_struct sonwav);
void Saisie_Nom();
void Recopier_wav(kwav_struct src, kwav_struct* dest);
void eventloop();
void print_infos(Window win);


int main(int argc, char *argv[])
{ 
  InitX11( 0, 0, L, H);
  kwav_Init(argc, argv);
  kfft_Init();

  sonwav.data = NULL;
  selwav.data = NULL;
  strcpy(nom, kwav_get_gfilename());
  
  print_infos(winInfo);
  eventloop();

  kwav_Release();
  kfft_Release();
  return EXIT_SUCCESS;
}


void InitX11(int posx, int posy, int largeur, int hauteur)
{
  if ((dpy = XOpenDisplay("")) == NULL){
      fprintf(stderr,"Cannot open display\n");
      exit(1);}

  ecran = DefaultScreen (dpy);
  racine = DefaultRootWindow (dpy);
  gc = DefaultGC (dpy, ecran);

  InitColor();
  XSetForeground(dpy, gc, jaune);
  InitWindows (posx, posy, largeur, hauteur);

  if ((fd = XLoadQueryFont(dpy, "fixed")) == NULL){
    fprintf(stderr, "Impossible de charger la police fixed\n");
    exit(1);}
  gcv.font = fd->fid;
  gcv.foreground = jaune;
  gcFont = XCreateGC(dpy, winInfo, GCFont | GCForeground, &gcv);

  XFlush (dpy);
}


void InitWindows(int posx, int posy, int largeur, int hauteur)
{
 
  /* fenetre mere */
  InitWindow (&winMere, &racine, posx, posy, largeur, hauteur, 
	      noir, gris, ExposureMask | ButtonPressMask | KeyPressMask); 
  XStoreName (dpy, winMere, "Kedit");

  /* fenetre panorama */
  InitWindow (&winPanorama, &winMere, posx+M, posy+H_BTN+(2*M), L_PAN, H_PAN, 
	      blanc, noir, ExposureMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask | Button1MotionMask);
  /* fenetre selection */
  InitWindow (&winSelection, &winMere, posx+M, posy+H_PAN+H_BTN+(3*M), L_SEL, H_SEL, 
	      blanc, noir, ExposureMask | ButtonPressMask | KeyPressMask | Button1MotionMask);
  /* fenetre sonogramme */
  InitWindow (&winSono, &winMere, posx+M, posy+H_PAN+H_SEL+H_BTN+(4*M), L_SNG, H_SNG,
  	      blanc, noir, ExposureMask | ButtonPressMask | KeyPressMask);
  /* fenetre fft */
  InitWindow (&winFft, &winMere, posx+L_SNG+(2*M), posy+H_PAN+H_SEL+H_BTN+(4*M), L_FFT, H_FFT,
  	      blanc, noir, ExposureMask | ButtonPressMask | KeyPressMask);
  /* fenetre info */
  InitWindow (&winInfo, &winMere, posx+L_SEL+(2*M), posy+H_BTN+(2*M), L_INF, H_INF,
  	      blanc, noir, ExposureMask | ButtonPressMask | KeyPressMask);

  /* fenetre bouton load */ 
  InitButton ("bouton.xpm", "LOAD", &winBLoad, &winMere, posx+M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton play */ 
  InitButton ("bouton.xpm", "PLAY", &winBPlay, &winMere, posx+L_BTN+2*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton rec. */ 
  InitButton ("bouton.xpm", "REC.", &winBRec, &winMere, posx+2*L_BTN+3*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton save */ 
  InitButton ("bouton.xpm", "SAVE", &winBSave, &winMere, posx+3*L_BTN+4*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton head */ 
  InitButton ("bouton.xpm", "HEAD", &winBHead, &winMere, posx+4*L_BTN+5*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton fft */ 
  InitButton ("bouton.xpm", "FFT", &winBFft, &winMere, posx+5*L_BTN+6*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  /* fenetre bouton quit */ 
  InitButton ("bouton.xpm", "QUIT", &winBQuit, &winMere, posx+6*L_BTN+7*M, posy+M, L_BTN, H_BTN,
	      blanc, blanc, ExposureMask | ButtonPressMask | ButtonReleaseMask);
}


void eventloop()
{
  int x0;

  while(1)
    {
      XNextEvent (dpy, &event);
      switch (event.type)
	{
	case ButtonPress:
	  switch (event.xbutton.button)
	    {
	    case Button1:
	      /* boutons */
	      if(event.xbutton.window == winBLoad)
		{
		  Xor(winBLoad);
		  Saisie_Nom();
		  kwav_Load(&sonwav, nom);
		  print_infos(winInfo);
		  Recopier_wav(sonwav, &selwav);
		  XClearWindow (dpy, winPanorama);
		  XClearWindow (dpy, winSelection);
		  Dessine_wav(winPanorama, sonwav);
		  Dessine_wav(winSelection, selwav);
		}
	      if(event.xbutton.window == winBPlay)
		{
		  Xor(winBPlay);
		  kwav_Play(selwav);
		}
	      if(event.xbutton.window == winBRec)
		{
		  Xor(winBRec);
		  free(sonwav.data);
		  kwav_Record(&sonwav, kwav_get_gsize());
		  print_infos(winInfo);
		  Recopier_wav(sonwav, &selwav);
		  XClearWindow (dpy, winPanorama);
		  XClearWindow (dpy, winSelection);
		  Dessine_wav(winPanorama, sonwav);
		  Dessine_wav(winSelection, selwav);
		}
	      if(event.xbutton.window == winBSave)
		{
		  Xor(winBSave);
		  Saisie_Nom();
		  kwav_Save(selwav, nom);
		}
	      if(event.xbutton.window == winBHead)
		{
		  Xor(winBHead);
		  kwav_Print_header (selwav.head);
		}
	      if (event.xbutton.window == winBFft)
		{
		  Xor(winBFft);
		  Dessine_sono(winSono, selwav);
		}
	      if(event.xbutton.window == winBQuit)
		{
		  Xor(winBQuit);
		}

	      /* fenetres */
	      if (event.xbutton.window == winFft)
		{
		  kfft_Analyse(selwav.data);
		  Dessine_fft(winFft, selwav);
		}
	      if(event.xbutton.window == winPanorama)
		{
		  x0 = event.xbutton.x;
		  XClearWindow (dpy, winPanorama);
		  Dessine_wav(winPanorama, sonwav);
		  XSetForeground(dpy, gc, vert);
		  XDrawLine(dpy, winPanorama, gc, x0, 0, x0, H_PAN);
		}
	      if(event.xbutton.window == winSelection)
		{
		  XClearWindow (dpy, winSelection);
		  Dessine_wav(winSelection, selwav);
		  XSetForeground(dpy, gc, vert);
		  XDrawLine(dpy, winSelection, gc, event.xbutton.x, 0, event.xbutton.x, H_SEL);
		  kfft_Analyse(selwav.data + (event.xbutton.x * selwav.head.data_length / selwav.head.byte_p_spl / L_SEL));
		  Dessine_fft(winFft, selwav);
		}
	      if (event.xbutton.window == winInfo)
		{
		  XClearWindow (dpy, winInfo);
		  print_infos(winInfo);
		}
	      break;

	    case Button2:
	      XClearWindow (dpy, event.xbutton.window);
	      break;
	    case Button3:
	      if(event.xbutton.window == winBQuit)
		{
		  Xor(winBQuit);
		}
	    }
	  break;

	case ButtonRelease:
	  switch (event.xbutton.button)
	    {
	    case Button1:
	      if(event.xbutton.window == winBLoad) 
		Xor(winBLoad);
	      if(event.xbutton.window == winBPlay) 
		Xor(winBPlay);
	      if(event.xbutton.window == winBRec) 
		Xor(winBRec);
	      if(event.xbutton.window == winBSave) 
		Xor(winBSave);
	      if(event.xbutton.window == winBHead) 
		Xor(winBHead);
	      if(event.xbutton.window == winBFft) 
		Xor(winBFft);
	      if(event.xbutton.window == winBQuit)
		{
		  Xor(winBQuit);
		  printf("au revoir! :-)\n");
		  XDestroyWindow (dpy, winMere);
		  XCloseDisplay (dpy);
		  return;
		}
	      if(event.xbutton.window == winPanorama)
		{
		  selwav.data = sonwav.data + (x0 * sonwav.head.data_length / sonwav.head.byte_p_spl / L_PAN);
		  selwav.head.data_length = (event.xbutton.x - x0)  * sonwav.head.data_length / L_PAN;
		  selwav.head.length = selwav.head.data_length + sizeof(selwav.head);
		  XClearWindow (dpy, winSelection);
		  Dessine_wav(winSelection, selwav);
		}
	      break;
	      
	    case Button3:
	      if(event.xbutton.window == winBQuit)
		{
		  Xor(winBQuit);
		  printf("Au revoir :-)\n");
		  XDestroyWindow (dpy, winMere);
		  XCloseDisplay (dpy);
		  return;
		}
	      break;
	    }
	  break;
	  
	case MotionNotify:
	    if(event.xbutton.window == winPanorama)
	      if(event.xmotion.state==272||event.xmotion.state==16||event.xmotion.state==256)
		{ 
		  XClearWindow (dpy, winPanorama);
		  XSetForeground (dpy, gc, vertFonce);
		  XFillRectangle (dpy, winPanorama, gc, x0, 0, event.xmotion.x-x0, H_PAN);
		  XSync (dpy, 0);
		  Dessine_wav(winPanorama, sonwav);
		}
	    if(event.xbutton.window == winSelection)
	      if(event.xmotion.state==272||event.xmotion.state==16||event.xmotion.state==256)
		{ 
		  XClearWindow (dpy, winSelection);
		  Dessine_wav(winSelection, selwav);
		  XSetForeground(dpy, gc, vert);
		  XDrawLine(dpy, winSelection, gc, event.xmotion.x, 0, event.xmotion.x, H_SEL);
		  XSync (dpy, 0);
		  kfft_Analyse(selwav.data + (event.xmotion.x * selwav.head.data_length / selwav.head.byte_p_spl / L_SEL));
		  Dessine_fft(winFft, selwav);
		}
	  break;
	}
    }
}


void print_infos(Window win)
{
  int saut = 13;
  char szVal[50];
  char szTxt[50];

  XClearWindow (dpy, win);

  XDrawString (dpy, win, gcFont, M, M+saut, "  Valeurs du fichier :", strlen ("  Valeurs du fichier :"));
  strcat (strcpy (szTxt, "nom    = "), nom);
  XDrawString (dpy, win, gcFont, M, M+3*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", sonwav.head.modus);
  strcat (strcpy (szTxt, "stereo = "), strcat (szVal, " voie(s)"));
  XDrawString (dpy, win, gcFont, M, M+4*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", sonwav.head.sample_fq);
  strcat (strcpy (szTxt, "freq   = "), strcat (szVal, " Hertz"));
  XDrawString (dpy, win, gcFont, M, M+5*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", sonwav.head.bit_p_spl);
  strcat (strcpy (szTxt, "bits   = "), strcat (szVal, " bits"));
  XDrawString (dpy, win, gcFont, M, M+6*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", sonwav.head.data_length);
  strcat (strcpy (szTxt, "taille = "), strcat (szVal, " octets"));
  XDrawString (dpy, win, gcFont, M, M+7*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_oct2ms(sonwav.head.data_length));
  strcat (strcpy (szTxt, "duree  = "), strcat (szVal, " millisecondes"));
  XDrawString (dpy, win, gcFont, M, M+8*saut, szTxt, strlen (szTxt));

  XDrawString (dpy, win, gcFont, M, M+11*saut, "  Valeurs d'environnement :", strlen ("  Valeurs d'environnement :"));
  strcat (strcpy (szTxt, "nom    = "), kwav_get_gfilename());
  XDrawString (dpy, win, gcFont, M, M+13*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_get_gstereo());
  strcat (strcpy (szTxt, "stereo = "), strcat (szVal, " voie(s)"));
  XDrawString (dpy, win, gcFont, M, M+14*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_get_gfreq());
  strcat (strcpy (szTxt, "freq   = "), strcat (szVal, " Hertz"));
  XDrawString (dpy, win, gcFont, M, M+15*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_get_gbits());
  strcat (strcpy (szTxt, "bits   = "), strcat (szVal, " bits"));
  XDrawString (dpy, win, gcFont, M, M+16*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_get_gsize());
  strcat (strcpy (szTxt, "taille = "), strcat (szVal, " octets"));
  XDrawString (dpy, win, gcFont, M, M+17*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_oct2ms(kwav_get_gsize()));
  strcat (strcpy (szTxt, "duree  = "), strcat (szVal, " millisecondes"));
  XDrawString (dpy, win, gcFont, M, M+18*saut, szTxt, strlen (szTxt));
  sprintf (szVal, "%d", kwav_get_gverbose());
  strcat (strcpy (szTxt, "print  = "), szVal);
  XDrawString (dpy, win, gcFont, M, M+19*saut, szTxt, strlen (szTxt));

}


void Dessine_wav(Window win, kwav_struct sonwav)
{
  int i;
  int nb_ech;
  int zero;
  double pas;
  unsigned int h;
  unsigned int l;
  double zoomy;
  KWAV_ECH8* ptr8;

  h = getHeight(win);
  l = getWidth(win);
  nb_ech = (sonwav.head.data_length) / (sonwav.head.byte_p_spl);
  pas = nb_ech / (double)l;
  zoomy = h / pow(2, sonwav.head.bit_p_spl);
  ptr8 = (KWAV_ECH8*)sonwav.data;

  XSetForeground(dpy, gc, rouge); /* horizontale */
  XDrawLine(dpy, win, gc, 0, h/2, l, h/2);
  XSetForeground(dpy, gc, jaune); /* signal */

  if(sonwav.head.bit_p_spl != 8)
    {
      zero = (h/2);
      for(i=0; (i+1)*pas<nb_ech; i++)
	XDrawLine(dpy, win, gc, 
		  i, sonwav.data[(int)(i*pas)] * zoomy + zero, 
		  i+1, sonwav.data[(int)((i+1)*pas)] * zoomy + zero);
    }
  else
    {
      zero = 0;
      for(i=0; (i+1)*pas<nb_ech; i++)
	XDrawLine(dpy, win, gc, 
		  i, ptr8[(int)(i*pas)] * zoomy + zero, 
		  i+1, ptr8[(int)((i+1)*pas)] * zoomy + zero);
    }
  XFlush(dpy);
}


void Saisie_Nom()
{
  char chaine[SIZE_NOM];

  printf("Nom du fichier (%s) : ", nom);
  fgets(chaine, SIZE_NOM, stdin);
  if(chaine[0] != '\n')
    {
      strcpy(nom, chaine);
      nom[strlen(nom)-1] = '\0';
    }
}


void Dessine_fft(Window win, kwav_struct sonwav)
{
  int i, maxbit, nbfreqs;
  unsigned int h;
  unsigned int l;

  h = getHeight(win);
  l = getWidth (win);
  nbfreqs = kfft_get_nbfreqs();
  maxbit = pow(2, sonwav.head.bit_p_spl) / 2;
  
  XSetForeground(dpy, gc, jaune);
  XClearWindow(dpy, win);
  for(i=0; i<h; i++)
      XDrawLine(dpy, win, gc, 0, h-i,
		((int)kfft_modules_reels[i*nbfreqs/h/2]*l/maxbit*5), h-i);
}


void Dessine_sono(Window win, kwav_struct sonwav)
{
  int i, j, amp, inc, nbfreqs;
  unsigned int h;
  unsigned int l;

  h = getHeight(win);
  l = getWidth (win);
  nbfreqs = kfft_get_nbfreqs();
  
  XClearWindow (dpy, win);
  inc = selwav.head.data_length / sonwav.head.byte_p_spl / l;

  for(i=0; i<l; i++)
    {
      kfft_Analyse(sonwav.data + i*inc);
      Dessine_fft(winFft, selwav);
      for(j=0; j<h; j++)
	{
	  amp = kfft_modules_reels[j*nbfreqs/h/2];
	  XSetForeground(dpy, gc, conversion((amp>512)?255:0, (amp>256)?255:amp, 0));
	  XDrawPoint(dpy, win, gc, i, h-j);
	  XFlush(dpy);
	}
    }
}


void InitColor()
{
  XAllocNamedColor (dpy, DefaultColormap(dpy, ecran), "red", &xcolor, &xcolor);  
  rouge = xcolor.pixel;
  XAllocNamedColor (dpy, DefaultColormap(dpy, ecran), "white", &xcolor, &xcolor);   
  blanc = xcolor.pixel;
  XAllocNamedColor (dpy, DefaultColormap(dpy, ecran), "black", &xcolor, &xcolor);   
  noir = xcolor.pixel;
  XAllocNamedColor (dpy, DefaultColormap(dpy, ecran), "blue", &xcolor, &xcolor);   
  bleu = xcolor.pixel;
  XAllocNamedColor( dpy, DefaultColormap(dpy, ecran), "SeaGreen", &xcolor, &xcolor);
  vertFonce = xcolor.pixel;
  XAllocNamedColor( dpy, DefaultColormap(dpy, ecran), "grey", &xcolor, &xcolor);
  gris = xcolor.pixel;
  XAllocNamedColor( dpy, DefaultColormap(dpy, ecran), "yellow", &xcolor, &xcolor);
  jaune = xcolor.pixel;
  XAllocNamedColor( dpy, DefaultColormap(dpy, ecran), "orchid", &xcolor, &xcolor);
  violet = xcolor.pixel;
  XAllocNamedColor( dpy, DefaultColormap(dpy, ecran), "aquamarine", &xcolor, &xcolor);
  vert = xcolor.pixel;
}


void InitWindow (Window *win, Window *mere, int posx, int posy, int largeur, int hauteur, int frontC, int backC, long event_mask)
{
  *win = XCreateSimpleWindow (dpy, *mere, posx, posy, largeur, hauteur, 0, frontC, backC);
  XChangeWindowAttributes (dpy, *win, CWBackingStore, &xswa);
  xswa.backing_store = Always;
  XSelectInput (dpy, *win, event_mask);
  XMapWindow (dpy, *win);
}


void InitButton (char* filename, char* nomBouton, Window *win, Window *mere, int posx, int posy, int largeur, int hauteur, int frontC, int backC, long event_mask)
{
  *win = XCreateSimpleWindow (dpy, *mere, posx, posy, largeur, hauteur, 0, frontC, backC);
  XChangeWindowAttributes (dpy, *win, CWBackingStore, &xswa);
  xswa.backing_store = Always;
  XSelectInput (dpy, *win, event_mask);
  XMapWindow (dpy, *win);
  XpmReadFileToPixmap (dpy, *win, filename, &pixmap, &pixmap, &attributes); 
  XCopyArea (dpy, pixmap, *win, gc, 0, 0, largeur, hauteur, 0, 0);
  XDrawString (dpy, *win, gc, 10, 20, nomBouton, strlen(nomBouton));
}


void Recopier_wav(kwav_struct src, kwav_struct* dest)
{
  dest->head = src.head;
  dest->data = src.data;
}


long conversion(int r,int v,int b)
{
  int newr, newv, newb;
  long couleur;
  newr= ((r*32)/256) << 11;
  newv= ((v*64)/256) << 5;
  newb= ((b*32)/256);
  couleur = newr + newv + newb;

  return couleur;
}


void separation(long couleur, int *r, int *v, int *b)
{
  *r = (couleur >> 11) << 3;
  *v = ((couleur >> 5) & 63) << 2;
  *b = (couleur & 31) << 3;
}


unsigned int getHeight(Window win)
{
  Window root_return;
  int x_return, y_return;
  unsigned int width_return, heigth_return;
  unsigned int border_width_return;
  unsigned int depth_return;

  XGetGeometry(dpy, win, &root_return, &x_return,
       &y_return, &width_return, &heigth_return, &border_width_return, &depth_return);

  return heigth_return;
}


unsigned int getWidth(Window win)
{
  Window root_return;
  int x_return, y_return;
  unsigned int width_return, heigth_return;
  unsigned int border_width_return;
  unsigned int depth_return;

  XGetGeometry(dpy, win, &root_return, &x_return,
       &y_return, &width_return, &heigth_return, &border_width_return, &depth_return);

  return width_return;
}


void Xor(Window winxor)
{
  XSetForeground (dpy, gc, blanc);
  XSetFunction (dpy, gc, GXxor);  //on pratique le xor
  XFillRectangle (dpy, winxor, gc, 0, 0, 100, 33);
  XSetFunction (dpy, gc, GXcopy); // on le rend transparent donc normal
  XFlush(dpy);
}
