#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "kwav.h"

#define _KFFT_DEFAULT_NBFREQS 1024


/** Gestion des nombres complexes **/

typedef struct complex {
  double reel;
  double imag; } complex;

#define kfft_complexmult(x, y, r) {\
r.reel = x.reel*y.reel - x.imag*y.imag;\
r.imag = x.reel*y.imag + x.imag*y.reel; }
 
#define kfft_complexplus(x, y, r) {\
r.reel = x.reel+y.reel;\
r.imag = x.imag+y.imag; }
 
#define kfft_complexmoins(x, y, r) {\
r.reel = x.reel-y.reel;\
r.imag = x.imag-y.imag; }

#define kfft_module(x) (sqrt(x.reel*x.reel + x.imag*x.imag))
#define phase(x)  (atan2(x.imag, x.reel))


/** Variables globales **/

complex *kfft_onde_complexe; /* signal de nbfreqs echantillons 
				ou spectre de nbfreqs frequences */
complex *kfft_ws;            /* racines niemes de 1 */
double  *kfft_modules_reels; /* amplitudes */
double  *kfft_phases_reels;  /* phases */
static unsigned int *invbin, *tmp;
int kfft_nbfreqs;
int inv;       /* 0 si fft directe, 1 si inverse */
int taille;    /* moitie du du nombre d'echantillons a traiter */
int anglebase; /* angle de rotation de l'onde generee par l'appel de papillon,
		  represente' par l'indice du complexe associe' dans le tableau
		  des racines niemes de l'unite' */


/** Prototypage **/

/* Algorithmes centraux */
void FFTiter(int nbfreqs);
void papillon(int k);

/* FFT directe (analyse en frequences) */
void fft();

/* FFT inverse (synthese additive) */
void fftinv();

/* Initialisations */
void transferer_signal(KWAV_ECH *s);
void racines_niemes_de_un (int n);
void inversbin(int k, int n); /* algo recursif */
void initinvbin();
void inverserbin();

/* Conversions */
void cartesiennes_vers_polaires();
void polaires_vers_cartesiennes();

/* Appels principaux */
void kfft_Init();
void kfft_Release();
void kfft_Analyse(KWAV_ECH *s);
void kfft_Synthese();
void kfft_set_nbfreqs (int nbfreqs);
int  kfft_get_nbfreqs ();
