/** Definitions **/

#define KWAV_FALSE 0
#define KWAV_TRUE 1

typedef signed short KWAV_ECH16; /* 16 bits par echantillon */
typedef unsigned char KWAV_ECH8; /*  8 bits par echantillon */
typedef KWAV_ECH16 KWAV_ECH;

typedef struct {        /* header for WAV-Files */
  char main_chunk[4];   /* 'RIFF' */
  int length;           /* length of file */
  char chunk_type[4];   /* 'WAVE' */
  char sub_chunk[4];    /* 'fmt' */
  int length_chunk;     /* length sub_chunk, always 16 bytes */
  short format;         /* always 1 = PCM-Code */
  short modus;          /* 1 = Mono, 2 = Stereo */
  int sample_fq;        /* Sample Freq */
  int byte_p_sec;       /* Data per sec */
  short byte_p_spl;     /* bytes per sample, 1=8 bit, 2=16 bit (mono)
		           2=8 bit, 4=16 bit (stereo) */
  short bit_p_spl;      /* bits per sample, 8, 12, 16 */
  char data_chunk[4];   /* 'data' */
  int data_length;      /* length of data */
} kwav_header;

typedef struct {
  kwav_header head;
  KWAV_ECH *data;
} kwav_struct;

int kwav_fdsp; /* descripteur de la carte son */



/** Prototypage **/

#ifdef __cplusplus
extern "C" {
int close (...);
#endif

extern void kwav_Init (int argc, char *argv[]); 
extern void kwav_Init_dsp (int freq_ech, int bits_par_ech, int en_stereo);
extern void kwav_Reconfig_dsp (int freq_ech, int bits_par_ech, int en_stereo);
extern void kwav_Release (void);
extern int kwav_ms2oct (int duree);
extern int kwav_oct2ms (int taille);
extern int kwav_ms2oct_from_header (int duree, kwav_header header);
extern int kwav_oct2ms_from_header (int taille, kwav_header header);

extern void kwav_Init_from_header (kwav_header header);
extern void kwav_Play (kwav_struct sonwav);
extern void kwav_Record (kwav_struct* psonwav, int nboctets);
extern void kwav_Save (kwav_struct sonwav, char nom_fichier[]);
extern void kwav_Load (kwav_struct* psonwav, char nom_fichier[]);
extern KWAV_ECH *kwav_Alloc (KWAV_ECH* pson, int nboctets);
extern void kwav_Free (KWAV_ECH* pson);
extern void kwav_Normalize (kwav_struct* wavptr);
extern void kwav_Print_header (kwav_header header);
extern kwav_header kwav_Create_header ();

extern int kwav_get_gstereo (void);
extern int kwav_get_gfreq (void);
extern int kwav_get_gbits (void);
extern int kwav_get_gsize (void);
extern int kwav_get_gblocksize (void);
extern int kwav_get_gverbose (void);
extern char* kwav_get_gfilename (void);
extern void kwav_set_gstereo (int stereo);
extern void kwav_set_gfreq (int freq);
extern void kwav_set_gbits (int bits);
extern void kwav_set_gsize (int taille);
extern void kwav_set_gverbose (int verbose);
extern void kwav_set_gfilename (char* filename);

#ifdef __cplusplus
}
#endif




