#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <linux/soundcard.h> 
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>


/*--------- definition des constantes et des variables globales ------------*/

typedef signed short KWAV_ECH16;  /* 16 bits par echantillon */
typedef unsigned char  KWAV_ECH8; /*  8 bits par echantillon */
typedef KWAV_ECH16 KWAV_ECH;

#define _KWAV_DEFAULT_STEREO 1
#define _KWAV_DEFAULT_FREQ_ECH 22050
#define _KWAV_DEFAULT_BITS (8 * sizeof(KWAV_ECH))
#define _KWAV_DEFAULT_TAILLE  (_KWAV_DEFAULT_FREQ_ECH * sizeof(KWAV_ECH) * _KWAV_DEFAULT_STEREO) /* 1 sec de son en octets */
#define _KWAV_DEFAULT_FILE_NAME "son.wav"
#define _KWAV_DEFAULT_WAVE_HEADER \
{ { 'R', 'I', 'F', 'F' }, \
  44, /* taille du header */ \
  { 'W', 'A', 'V', 'E' }, \
  { 'f', 'm', 't', ' ' }, \
  16, \
  1, \
  _KWAV_DEFAULT_STEREO, \
  _KWAV_DEFAULT_FREQ_ECH, \
  _KWAV_DEFAULT_FREQ_ECH * _KWAV_DEFAULT_STEREO * sizeof(KWAV_ECH), \
  _KWAV_DEFAULT_STEREO * sizeof(KWAV_ECH), sizeof(KWAV_ECH) * 8, \
  { 'd', 'a', 't', 'a' }, \
  0 }

#define KWAV_FALSE 0
#define KWAV_TRUE 1

int kwav_fdsp;          /* file descriptor de la carte son */
int gstereo;            /* stereo globale */
int gfreq;              /* frequence globale */
int gbits;              /* resolution globale */
int gtaille;            /* taille globale en octet */
int gtaillebloc;        /* taille des blocs de la carte son */
char gnom_fichier[256]; /* nom global */
int gverbose;             /* drapeau global pour les printf */


/*---------------- necessaire a la generation d'un fichier WAV -----------*/

typedef struct {        /* header for WAV-Files */
  char main_chunk[4];   /* 'RIFF' */
  int length;           /* length of file */
  char chunk_type[4];   /* 'WAVE' */
  char sub_chunk[4];    /* 'fmt' */
  int length_chunk;     /* length sub_chunk, always 16 bytes */
  short format;         /* always 1 = PCM-Code */
  short modus;          /* 1 = Mono, 2 = Stereo */
  int sample_fq;        /* Sample Freq */
  int byte_p_sec;       /* Data per sec */
  short byte_p_spl;     /* bytes per sample, 1=8 bit, 2=16 bit (mono)
			   2=8 bit, 4=16 bit (stereo) */
  short bit_p_spl;      /* bits per sample, 8, 12, 16 */
  char data_chunk[4];   /* 'data' */
  int data_length;      /* length of data */
} kwav_header;

typedef struct {
  kwav_header head;
  KWAV_ECH *data;
} kwav_struct;


/*------------------------ prototypage ---------------------------------*/

void kwav_Init (int argc, char *argv[]); 
void kwav_Init_dsp (int freq_ech, int bits_par_ech, int en_stereo);
void kwav_Reconfig_dsp (int freq_ech, int bits_par_ech, int en_stereo);
void kwav_Release ();
int kwav_ms2oct (int duree);
int kwav_oct2ms (int taille);
int kwav_ms2oct_from_header (int duree, kwav_header header);
int kwav_oct2ms_from_header (int taille, kwav_header header);

void kwav_Init_from_header (kwav_header header);
void kwav_Play (kwav_struct sonwav);
void kwav_Record (kwav_struct* psonwav, int nboctets);
void kwav_Save (kwav_struct sonwav, char nom_fichier[]);
void kwav_Load (kwav_struct* psonwav, char nom_fichier[]);
KWAV_ECH *kwav_Alloc (KWAV_ECH* pson, int nboctets);
void kwav_Free (KWAV_ECH* pson);
void kwav_Normalize (kwav_struct* wavptr);
void kwav_Print_header (kwav_header header);
kwav_header kwav_Create_header ();

int kwav_get_gstereo ();
int kwav_get_gfreq ();
int kwav_get_gbits ();
int kwav_get_gsize ();
int kwav_get_gblocksize ();
int kwav_get_gverbose ();
char* kwav_get_gfilename ();
void kwav_set_gstereo (int stereo);
void kwav_set_gfreq (int freq);
void kwav_set_gbits (int bits);
void kwav_set_gsize (int taille);
void kwav_set_gverbose (int verbose);
void kwav_set_gfilename (char* filename);
