/*******************************************************/
/* FFT iterative, code de Daniel Goossens, decembre 99 */
/* Necessite la bibliotheque kwav de Karim Barkati     */
/*******************************************************/

#include "kfftlib.h"


void kfft_Init()
{
  kfft_set_nbfreqs (_KFFT_DEFAULT_NBFREQS);

  (void*)kfft_onde_complexe = malloc(sizeof(complex) * kfft_nbfreqs);
  (void*)kfft_modules_reels = malloc(sizeof(double) * kfft_nbfreqs);
  (void*)kfft_phases_reels  = malloc(sizeof(double) * kfft_nbfreqs);

  (void*)invbin = malloc(sizeof(unsigned int) * kfft_nbfreqs);
  (void*)kfft_ws = malloc(sizeof(complex) * kfft_nbfreqs);

  racines_niemes_de_un (kfft_nbfreqs); /* les nbfreqs racines nbfreqs-iemes de 1 */
}


void kfft_Release()
{
  free(kfft_onde_complexe);
  free(kfft_modules_reels);
  free(kfft_phases_reels);

  free(invbin);
  free(kfft_ws);
}


void kfft_set_nbfreqs (int nbfreqs){
  kfft_nbfreqs = nbfreqs; }


int kfft_get_nbfreqs (){
  return kfft_nbfreqs; }


/** Appels principaux **/

void kfft_Analyse(KWAV_ECH *s)
{
  int i;

  initinvbin(); /* initialisation inversion binaire */
  /* s est un signal contenant au moins kfft_nbfreqs echantillons */
  transferer_signal(s); /* remplit kfft_onde_complexe avec les nbfreqs premiers echantillons du signal s. Ca fait aussi l'inversion binaire. kfft_onde_complexe est melangee */
  
  if(kwav_get_gverbose() == KWAV_TRUE)
    {
      for(i=0; i<kfft_nbfreqs; i++)
	printf("kfft_onde_complexe[%d].reel = %f\n", i, 
	       kfft_onde_complexe[i].reel);
      printf("appel fft()...\n");
    }

  fft(); /* defait l'inversion binaire. kfft_onde_complexe n'est plus melangee */
  
  if(kwav_get_gverbose() == KWAV_TRUE)
    printf("->fin fft()\n");

  cartesiennes_vers_polaires(); /* remplit kfft_phases_reels et kfft_modules_reels a partir de kfft_onde_complexe */
}


void kfft_Synthese()
{
  /* suppose kfft_phases_reels et kfft_modules_reels deja remplis */
  polaires_vers_cartesiennes(); /* remplit kfft_onde_complexe a partir de kfft_phases_reels et kfft_modules_reels */
  inverserbin(); /* refait l'inversion binaire. kfft_onde_complexe est melangee */

  cartesiennes_vers_polaires();
  fftinv(); /* defait l'inversion binaire. kfft_onde_complexe n'est plus melangee */
}


/** Fonctions internes **/

/* appel: papillon(indice-case-depart). 
   Presuppose les globales taille 
   (moitie de la portion du tableau a traiter) 
   et anglebase (increment du parcours du tableau 
   de racines niemes de l'unite kfft_ws) */

void papillon(int k)
{
  /* k = indice du 1er echantillon a traiter */
  complex angle, t;
  int a = 0; /* nb d'anglebase dans l'angle courant */
  int i;
  int fin = k + taille; /* indice de depart + (nb d'ech traiter/2) */
  
  for(i=k; i<fin; i++)
    {
      if(inv)
	angle = kfft_ws[a];
      else
	angle = kfft_ws[(a==0 ? 0 : kfft_nbfreqs-a)];
      
      kfft_complexmult(angle, kfft_onde_complexe[taille+i], t);
      kfft_complexmoins(kfft_onde_complexe[i], t, kfft_onde_complexe[taille+i]);
      kfft_complexplus(kfft_onde_complexe[i], t, kfft_onde_complexe[i]);

      a += anglebase; /* tourner d'un anglebase */
    }
}


/** Algorithme central **/

void FFTiter(int nbfreqs)
{
  /* on travaille sur les nbfreqs complexes ranges dans onde_complex */
  int i, n;
  
  for(i=0; i<nbfreqs; i+=2) /* cas de base direct sans multiplication */
    {
      complex c;
      c = kfft_onde_complexe[i+1];
      kfft_complexmoins(kfft_onde_complexe[i], c, kfft_onde_complexe[i+1]);
      kfft_complexplus(kfft_onde_complexe[i], c, kfft_onde_complexe[i]);
    }
 
  anglebase = nbfreqs>>2; //division par 4
  taille = 2; /* moitie de n */

  for(n=4; n<nbfreqs; taille=n, n<<=1, anglebase>>=1)
    for(i=0; i<nbfreqs; i+=n)
      papillon(i);

  taille = nbfreqs>>1; // division par 2
  anglebase = 1; /* onde de plus grande frequence. de 1 en 1. */
  papillon(0);
 
} 


/* FFT directe (analyse en frequences) */
void fft()
{
  int i;
  inv = KWAV_FALSE;
  FFTiter(kfft_nbfreqs);
  /* diviser par kfft_nbfreqs, juste pour rendre inversible */
  for(i=0; i<kfft_nbfreqs; i++)
    {
      kfft_onde_complexe[i].reel /= (double)kfft_nbfreqs;
      kfft_onde_complexe[i].imag /= (double)kfft_nbfreqs;
    }
}


/* FFT inverse (synthese additive) */
void fftinv()
{
  inv = KWAV_TRUE;
  FFTiter(kfft_nbfreqs);
}


/** Initialisations **/

void transferer_signal(KWAV_ECH *s)
{
  /* s contient des entiers entre 0 et MAXAMPLI */
  int i;

   for(i=0; i<kfft_nbfreqs; i++) 
    { 
      /* dans onde complexe on a les 1024 premieres valeurs
	 du son brut melanges */
      kfft_onde_complexe[invbin[i]].reel = s[i];
      kfft_onde_complexe[i].imag = 0;
    }
}


void racines_niemes_de_un (int n)
{
  int i;
  /* remplit de kfft_ws[0] a kfft_ws[n-1] avec les n racines niemes de 1 */
  /* c'est aussi l'onde de frequence la + petite (fech/n) sur une
     periode (n echantillons) */
  double angle_mini = 2 * M_PI / ((double) n);
  /* il faut successivement ajouter des angles_mini jusqu'a faire le
     tour du cercle. Chaque angle A est represente par ses coordonnees
     cartesiennes cos(A) et sin(A), par le complexe cos(A)+i*sin(A), et
     la somme de deux angles est la multiplication de leurs complexes */
  complex w;
  /* angle correspondant a 1/nieme de cercle: */
  w.reel = cos(angle_mini);
  w.imag = sin(angle_mini);

  /* position de depart sur le cercle */
  kfft_ws[0].reel = 1; 
  kfft_ws[0].imag = 0;

  /* on tourne n fois pour faire le tour du cercle */
  for(i=1; i<n; i++)
    kfft_complexmult(kfft_ws[i-1], w, kfft_ws[i]);
}


void inversbin(int k, int n) /* algo recursif */
{
  /* 1. perfect unshuffle entre les cases k et k+n-1 */
  int i, h, j;
  h = n>>1;

  if(n==2)
    return;
  
  for(i=0; i<h; i++)
    {
      j = k + 2 * i;
      tmp[i] = invbin[j];
      tmp[i+h] = invbin[j+1];
    }
  for(i=0; i<n; i++)
    invbin[k+i] = tmp[i]; //invbin tablau desordonne

  /* 2. appels recursifs sur les deux moities */
  inversbin(k, h);
  inversbin(k+h, h);
}


void initinvbin()
{
  int i;

  (void*)tmp = malloc(sizeof(unsigned int) * kfft_nbfreqs);
 
  for(i=0; i<kfft_nbfreqs; i++)
    invbin[i] = i;

  inversbin(0, kfft_nbfreqs);
  free(tmp);
}


void inverserbin()
{
  int i;
  complex *tmpx;
  (void*)tmpx = malloc(sizeof(complex) * kfft_nbfreqs);
  for(i=0; i<kfft_nbfreqs; i++)
    tmpx[invbin[i]] = kfft_onde_complexe[i];
  for(i=0; i<kfft_nbfreqs; i++)
     kfft_onde_complexe[i] = tmpx[i];
  free(tmpx);
}


/** Conversions **/

void cartesiennes_vers_polaires()
{
  /* exprime chaque vecteur complexe a+ib de onde_complexe
     en coordonnees polaires:
     module = racine(a*a+b*b)
     phase = arctg(b/a) */
  unsigned int temps;
  
  for(temps=0; temps<kfft_nbfreqs; temps++)
    {
      /* amplitudes */
      kfft_modules_reels[temps] = kfft_module (kfft_onde_complexe[temps]);
      /* phases possiblement aleatoires */
      kfft_phases_reels [temps] = phase (kfft_onde_complexe[temps]); 
    }
}


void polaires_vers_cartesiennes()
{
  /* traduction module+phase en un complexe:
     module*cos(phase) + i*phase*sin(phase) */
  unsigned int temps;
  
  for(temps=0; temps<kfft_nbfreqs; temps++)
    {
      kfft_onde_complexe[temps].reel =  /* modules = amplitudes */
	kfft_modules_reels[temps] * cos(kfft_phases_reels[temps]);
      kfft_onde_complexe[temps].imag =  /* phases = possiblement aleatoires */
	kfft_modules_reels[temps] * sin(kfft_phases_reels[temps]);
    }
}
