/*******************************************************************/
/*         Bibliotheque de fonctions audio :                       */
/*             Gestion de la carte son                             */
/*             Gestion du format wav                               */
/*             Gestion de la ligne de commande                     */
/*                       Karim Barkati, 2000                       */
/*******************************************************************/

#include "kwavlib.h"


void kwav_Init (int argc, char *argv[])
{
  int i;
  char wav[] = ".wav";
  char ext[4];

  /* Parametrage par defaut */
  kwav_set_gfreq (_KWAV_DEFAULT_FREQ_ECH);
  kwav_set_gbits (8 * sizeof(KWAV_ECH));
  kwav_set_gstereo (_KWAV_DEFAULT_STEREO);
  kwav_set_gsize (_KWAV_DEFAULT_TAILLE);
  kwav_set_gverbose (KWAV_FALSE);
  kwav_set_gfilename (_KWAV_DEFAULT_FILE_NAME);

  /* Verification des options */
  if(argc > 1 && argv[1] != '\0')
    {
      for(i=1; i<argc; i++)
	if(strcmp(argv[i], "-v")==0 || strcmp(argv[i], "--verbose")==0)
	  kwav_set_gverbose (KWAV_TRUE);
	
      if(gverbose == KWAV_TRUE)
	printf("\tOption(s) demandee(s): ");
      
      for(i=1; i<argc; i++)
	{
	  if(gverbose == KWAV_TRUE)
	    printf("%s ", argv[i]); /* simple echo */
	  
	  if(strcmp(argv[i], "-s")==0)
	    kwav_set_gstereo (atoi(argv[i+1]));   /* m=1 et s=2 */
	  if(strcmp(argv[i], "-f")==0)
	    kwav_set_gfreq (atoi(argv[i+1]));     /* en hertz */
	  if(strcmp(argv[i], "-b")==0){
	    kwav_set_gbits (atoi(argv[i+1]));
	    if (gbits != 8) gbits = 16;} /* seulement 8 ou 16 */
	  if(strcmp(argv[i], "-d")==0)
	    kwav_set_gsize (kwav_ms2oct(atoi(argv[i+1]))); /* en ms */
	  if(strcmp(argv[i], "-n")==0)
	    kwav_set_gfilename (argv[i+1]);
	  if(strcmp(argv[i], "-h")==0 || strcmp(argv[i],"--help")==0)
	    {
	      printf("\n"
		     "<fichier.wav>\n"
		     "-b <resolution en bits par echantillon>, 8 ou 16\n"
		     "-d <duree en millisecondes>\n"
		     "-f <frequence d'echantillonnage>, 11025 ou 22050 ou 44100\n"
		     "-n <nom du fichier>\n"
		     "-s <stereophonie>, 1 ou 2\n"
		     "-v ou --verbose\n"
		     "-h ou --help\n");
	      exit(0);
	    }
	  /* recherche de l'extension */
	  strcpy(ext, argv[i]+strlen(argv[i])-4);
	  if(strcmp(ext, wav)==0)
	    strcpy(gnom_fichier, argv[i]);
	}
      if(gverbose == KWAV_TRUE)
	printf("\n");
    }

  if(gverbose == KWAV_TRUE) {
    printf("\tValeurs par defaut :\n");
    printf("stereo = %d voie\n", gstereo);
    printf("freq   = %d Hertz\n", gfreq);
    printf("bits   = %d bits\n", gbits);
    printf("duree  = %d ms\n", kwav_oct2ms(gtaille));
    printf("nom    = %s\n", gnom_fichier);
  }

  /* Parametrage du peripherique */
  kwav_Init_dsp(gfreq, gbits, gstereo);
}


void kwav_Init_dsp (int freq_ech, int bits_par_ech, int en_stereo)
{
  /* ouvre un fichier dsp en lecture et en ecriture */
  if ((kwav_fdsp = open("/dev/dsp", O_RDWR))<0){
      perror("kwav: Ouverture du fichier /dev/dsp");
      return;}

  /* si le son n'est pas lu a la bonne vitesse,
     decommenter la ligne suivante et 
     commenter celle d'apres */

  //if(en_stereo==2) en_stereo=0; /* 0 ou 1 pour la carte */
  en_stereo--;

  /* kwav_fdsp est suppose ouvert */
  ioctl(kwav_fdsp, SNDCTL_DSP_STEREO, &en_stereo);
  ioctl(kwav_fdsp, SNDCTL_DSP_SPEED, &freq_ech);
  ioctl(kwav_fdsp, SNDCTL_DSP_SAMPLESIZE, &bits_par_ech);
  ioctl(kwav_fdsp, SNDCTL_DSP_GETBLKSIZE, &gtaillebloc);

  if(gverbose == KWAV_TRUE) {
    printf("\tkwav_Init_dsp :\n");
    printf("stereophonie                = %d voie(s)\n", en_stereo+1);
    printf("frequence d'echantillonnage = %d hertz\n", freq_ech);
    printf("taille d'echantillon        = %d bits\n", bits_par_ech);
    printf("taille de bloc              = %d octets\n", gtaillebloc);
  }
}


void kwav_Init_from_header (kwav_header header)
{
  int stereo, freq, bits;

  stereo = header.modus;
  freq   = header.sample_fq;
  bits   = header.bit_p_spl;

  kwav_Reconfig_dsp(freq, bits, stereo);
}


void kwav_Reconfig_dsp (int freq_ech, int bits_par_ech, int en_stereo)
{
  /* kwav_fdsp est suppose ouvert */
  //if(en_stereo==2) en_stereo=0; /* 0 ou 1 pour la carte */
  en_stereo--;
  ioctl(kwav_fdsp, SNDCTL_DSP_STEREO, &en_stereo);
  ioctl(kwav_fdsp, SNDCTL_DSP_SPEED, &freq_ech);
  ioctl(kwav_fdsp, SNDCTL_DSP_SAMPLESIZE, &bits_par_ech);

  if(gverbose == KWAV_TRUE) {
    printf("\tkwav_Reconfig_dsp :\n");
    printf("stereophonie                = %d voie(s)\n", en_stereo+1);// + 1);
    printf("frequence d'echantillonnage = %d Hertz\n", freq_ech);
    printf("taille d'echantillon        = %d bits\n", bits_par_ech);
  }
}


kwav_header kwav_Create_header () /* en fct des globales d'environnement */
{
  kwav_header new_header = _KWAV_DEFAULT_WAVE_HEADER;

  new_header.length = sizeof(new_header) + gtaille;
  new_header.modus = gstereo;
  new_header.sample_fq = gfreq;
  new_header.byte_p_sec = gfreq * gstereo * (gbits/8);
  new_header.byte_p_spl = gstereo * (gbits/8);
  new_header.bit_p_spl = gbits;
  new_header.data_length = gtaille;

  return new_header;
}


void kwav_Record (kwav_struct *psonwav, int nboctets)
{
  psonwav->head = kwav_Create_header();
  psonwav->data = kwav_Alloc (psonwav->data, nboctets);

  if(gverbose == KWAV_TRUE)
    printf("enregistrer... \n");
  
  ioctl(kwav_fdsp, SNDCTL_DSP_SYNC, 0);
  if (read(kwav_fdsp, psonwav->data, nboctets)<0){
    perror("kwav: Read du dsp");
    return;}
  ioctl(kwav_fdsp, SNDCTL_DSP_SYNC, 0);
  
  if(gverbose == KWAV_TRUE)
    printf("->fin enregistrer\n");
  
  psonwav->head.length = sizeof(psonwav->head) + nboctets;
  psonwav->head.data_length = nboctets;
}


void kwav_Play (kwav_struct sonwav)
{
  kwav_Init_from_header(sonwav.head);
  
  if(gverbose == KWAV_TRUE)
    printf("jouer...\n");
  ioctl(kwav_fdsp, SNDCTL_DSP_SYNC, 0);
  if (write(kwav_fdsp, sonwav.data, sonwav.head.data_length)<0){
    perror("kwav: Write sur dsp");
    return;}
  ioctl(kwav_fdsp, SNDCTL_DSP_SYNC, 0);
  if(gverbose == KWAV_TRUE)
    printf("->fin jouer\n");
}


void kwav_Save (kwav_struct sonwav, char nom_fichier[])
{
  int i;
  int fdwav;
  char wav[] = ".wav";

  /* recherche de l'extension, sinon ajout */
  for(i=0; i<4; i++)
    if(nom_fichier[strlen(nom_fichier)-4+i] != wav[i])
      strcat(nom_fichier, wav);
  
  /* ecrit un fichier son.wav sur le disque */
  if((fdwav = open(nom_fichier, O_WRONLY | O_CREAT, S_IWRITE | S_IREAD)) == -1){
    perror("kwav: Ouverture du fichier wav");
    return;}

  if(gverbose == KWAV_TRUE){
    printf("ecriture du fichier \"%s\"... \n", nom_fichier);
    printf(" header...");}
  write(fdwav, &sonwav.head, sizeof(sonwav.head));
  if(gverbose == KWAV_TRUE)
    printf(" data...\n");
  write(fdwav, sonwav.data, sonwav.head.data_length);
  if(gverbose == KWAV_TRUE)
    printf("->fin d'ecriture du fichier \"%s\"\n", nom_fichier);

  if((close(fdwav)) == -1){
    perror("kwav: Fermeture du fichier wav");
    return;}
}


void kwav_Print_header (kwav_header header)
{
  char tmpstr[5] = "0000\0";

  memcpy(tmpstr, header.main_chunk, 4);
  printf("{ main_chunk   = '%s'\n", tmpstr);
  printf("{ length       = %d octets\n", header.length);
  memcpy(tmpstr, header.chunk_type, 4);
  printf("{ chunk_type   = '%s'\n", tmpstr);
  tmpstr[3]='\0'; memcpy(tmpstr, header.sub_chunk, 3);
  printf("{ sub_chunk    = '%s'\n", tmpstr);
  printf("{ length_chunk = %d octets\n", header.length_chunk);
  printf("{ format       = %d %s\n", header.format, header.format==1 ? "(PCM-code)" : "");
  printf("{ modus        = %d (%s)\n", header.modus, header.modus==1 ? "mono" : "stereo");
  printf("{ sample_fq    = %d echantillons/s\n", header.sample_fq);
  printf("{ byte_p_sec   = %d octets/s\n", header.byte_p_sec);
  printf("{ byte_p_spl   = %d octets/echantillon\n", header.byte_p_spl);
  printf("{ bit_p_spl    = %d bits/echantillon\n", header.bit_p_spl);
  memcpy(tmpstr, header.data_chunk, 4);
  printf("{ data_chunk   = '%s'\n", tmpstr);
  printf("{ data_length  = %d octets\n", header.data_length);
}


void kwav_Load (kwav_struct* psonwav, char nom_fichier[])
{
  int fdwav;
  
  /* ouverture du fichier */
  if((fdwav = open(nom_fichier, O_RDONLY)) == -1){
    perror("kwav: Ouverture du fichier en lecture");
    return;}

  if(gverbose == KWAV_TRUE){
    printf("lecture du fichier \"%s\": \n", nom_fichier);
    printf("\tlecture header...\n");}

  read (fdwav, &psonwav->head.main_chunk, 4*sizeof(char) );    /* 'RIFF' */
  read (fdwav, &psonwav->head.length,       sizeof(int)  );    /* length of file */
  read (fdwav, &psonwav->head.chunk_type, 4*sizeof(char) );    /* 'WAVE' */
  read (fdwav, &psonwav->head.sub_chunk,  4*sizeof(char) );    /* 'fmt' */
  read (fdwav, &psonwav->head.length_chunk, sizeof(int)  );    /* 16 bytes */
  read (fdwav, &psonwav->head.format,       sizeof(short));    /* 1=PCM-Code */
  read (fdwav, &psonwav->head.modus,        sizeof(short));    /* 1=Mono, 2=Stereo */
  read (fdwav, &psonwav->head.sample_fq,    sizeof(int)  );    /* Sample Freq */
  read (fdwav, &psonwav->head.byte_p_sec,   sizeof(int)  );    /* Data per sec */
  read (fdwav, &psonwav->head.byte_p_spl,   sizeof(short));    /* 1, 2, 3, 4 */
  read (fdwav, &psonwav->head.bit_p_spl,    sizeof(short));    /* 8, 12, 16 */
  read (fdwav, &psonwav->head.data_chunk, 4*sizeof(char) );    /* 'data' */
  read (fdwav, &psonwav->head.data_length,  sizeof(int)  );    /* length of data */

  if(gverbose == KWAV_TRUE){
    kwav_Print_header (psonwav->head);   
    printf("\tlecture  data...\n");}

  psonwav->data = kwav_Alloc (psonwav->data, psonwav->head.data_length);
  read (fdwav, psonwav->data, psonwav->head.data_length);  /* chargement du son lui-meme */
  
  if(gverbose == KWAV_TRUE)
    printf("->fin lecture \"%s\"\n", nom_fichier);
  if((close(fdwav)) == -1){
    perror("kwav: Fermeture du fichier wav");
    return;}
}


KWAV_ECH *kwav_Alloc (KWAV_ECH* pson, int nboctets)
{
  if(gverbose == KWAV_TRUE)
    printf("allouer... \n");
  pson = malloc (nboctets); /* allocation en octets */
  if (pson==NULL){
    perror("kwav: Allocation du pointeur");
    return pson;}
  if(gverbose == KWAV_TRUE)
    printf("->fin allouer\n");
  return pson;
}


void kwav_Free (KWAV_ECH* pson)
{
  free(pson);
}


void kwav_Release ()
{
  if((close(kwav_fdsp)) == -1){
    perror("kwav: Fermeture du dsp");
    return;}
}


/* converti des millisecondes en octets a partir des globales */
int kwav_ms2oct (int duree)
{
  int taille;
  taille = duree * gfreq * gstereo * (gbits/8) * 0.001;
  return taille;
}


/* converti des millisecondes en octets a partir des globales */
int kwav_oct2ms (int taille)
{
  int duree;
  duree = taille * 1000 / (gfreq * gstereo * (gbits/8));
  return duree;
}


/* converti des millisecondes en octets a partir du header */
int kwav_ms2oct_from_header (int duree, kwav_header header)
{
  int taille;
  taille = duree * header.sample_fq * header.byte_p_spl * header.modus * 0.001;
  return taille;
}


/* converti des millisecondes en octets a partir des globales */
int kwav_oct2ms_from_header (int taille, kwav_header header)
{
  int duree;
  duree = taille * 1000 / (header.sample_fq * header.byte_p_spl * header.modus);
  return duree;
}


/* Accesseurs des variables globales */
int kwav_get_gstereo (){
  return gstereo; }

int kwav_get_gfreq (){
  return gfreq; }

int kwav_get_gbits (){
  return gbits; }

int kwav_get_gsize (){
  return gtaille; }

int kwav_get_gblocksize (){
  return gtaillebloc; }

int kwav_get_gverbose (){
  return gverbose; }

char* kwav_get_gfilename (){
  return gnom_fichier; }


/* Modificateurs des variables globales */
void kwav_set_gstereo (int stereo){
  gstereo = stereo; }

void kwav_set_gfreq (int freq){
  gfreq = freq; }

void kwav_set_gbits (int bits){
  gbits = bits; }

void kwav_set_gsize (int taille){
  gtaille = taille; }

void kwav_set_gverbose (int verbose){
  gverbose = verbose; }

void kwav_set_gfilename (char* filename){
  strcpy(gnom_fichier, filename); }


/* Fonction utilitaire de mise au maximum */
void kwav_Normalize (kwav_struct* wavptr)
{
  int i, ampmax, ampmin, imax, imin, nb_ech, maxbit;
  double coeff;
  KWAV_ECH8* ptr8;

  if (wavptr->head.bit_p_spl != 8 && wavptr->head.bit_p_spl != 16)
    {
      printf("Ce fichier est en %d bits, mais ce programme ne supporte que 8 ou 16 bits par echantillon\n", wavptr->head.bit_p_spl);
      return;
    }

  nb_ech = (wavptr->head.data_length) / (wavptr->head.byte_p_spl);
  maxbit = pow(2, wavptr->head.bit_p_spl) / 2;
  ampmax = 0; imax = 0;
  ampmin = 0; imin = 0;

  /* recherche de l'amplitude maximum */
  if(wavptr->head.bit_p_spl == 16)
    {
      for(i=0; i<nb_ech; i++)
	{
	  if(wavptr->data[i] > ampmax)
	    {
	      ampmax = wavptr->data[i];
	      imax = i;
	    }
	  if(wavptr->data[i] < ampmin)
	    {
	      ampmin = wavptr->data[i];
	      imin = i;
	    }
	}
    }
  else
    {
      ptr8 = (KWAV_ECH8*)wavptr->data;
      for(i=10; i<nb_ech; i++)
	{
  	  if(ptr8[i] > ampmax)
  	    {
  	      ampmax = ptr8[i];
 	      imax = i;
  	    }
 	  if(ptr8[i] < ampmin)
 	    {
  	      ampmin = ptr8[i];
 	      imin = i;
  	    }
	}
    }
  
  /* calcul du coeff en fonction d'ampmax ou ampmin */
  coeff = ((double) maxbit / (ampmax > (-ampmin) ? ampmax : -ampmin ));
  
  if(gverbose == KWAV_TRUE)
    printf("maxbit = %d, ampmax[%d] = %d, ampmin[%d] = %d, coeff = %f\n", 
	   maxbit, imax, ampmax, imin, ampmin, coeff);
  
  /* multiplication de chaque echantillon */
  if(coeff == 1)
    printf("Son deja au maximum\n");
  else
    {
      for(i=0; i<nb_ech; i++)
	{
	  if(wavptr->head.bit_p_spl==16)
	    wavptr->data[i] *= coeff;
	  else
	    ptr8[i] *= coeff;
	}
    }
  printf("Coefficient = %.2f\n", coeff);
}
